/*
    ES Knot
    Copyright 2005, 2006 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsj
    martin@martin.st
*/

#include "knot.h"
#include <GLES/gl.h>
#ifndef NO_EGL
#include <GLES/egl.h>
#endif
#include <math.h>
#include <stdio.h>
#include <string.h>

void debug(const char *str);

#ifndef M_PI
#define M_PI 3.141592654
#endif

#ifndef GL_OES_VERSION_1_1
#ifdef USE_OES_1_1
#undef UES_OES_1_1
#endif
#endif


#ifdef __WINSCW__
static void foo() {
	// workaround for strange cw compiler bug
	// otherwise we get multiply defined symbols regarding
	// operator new, if the first use of new is new[]
	int *i = new int();
}
#endif


#define printGLErrors() do { int err = glGetError(); if (err) printf("%d: %x\n", __LINE__, err); } while (0)

#define T_RAD1 2.28
#define T_RAD2 1.00
#define RADIUS F(0.7)
#define BUMPSIZE 0.20
#define OUTLINEADD 0.11
#define BUMPS 7

#define SUB3(a, b)		\
	do {			\
		a[0] -= b[0];	\
		a[1] -= b[1];	\
		a[2] -= b[2];	\
	} while (0);
#ifdef USING_FIXED
#define NORMALIZE3(v)							\
	do {								\
		float len = sqrt(float(MULT(v[0], v[0]) + MULT(v[1], v[1]) + MULT(v[2], v[2]))/(1<<16));	\
		v[0] = TYPE(v[0] / len);				\
		v[1] = TYPE(v[1] / len);				\
		v[2] = TYPE(v[2] / len);				\
	} while (0);
#else
#define NORMALIZE3(v)							\
	do {								\
		TYPE len = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);	\
		v[0] /= len;						\
		v[1] /= len;						\
		v[2] /= len;						\
	} while (0);
#endif
#define COPY3(a, b)		\
	do {			\
		a[0] = b[0];	\
		a[1] = b[1];	\
		a[2] = b[2];	\
	} while (0);

static void cross3(TYPE* out, const TYPE* v1, const TYPE* v2) {
	out[0] = MULT(v1[1], v2[2]) - MULT(v1[2], v2[1]);
	out[1] = MULT(v1[2], v2[0]) - MULT(v1[0], v2[2]);
	out[2] = MULT(v1[0], v2[1]) - MULT(v1[1], v2[0]);
}

void Knot::polar(float pos, TYPE *vect, int P, int Q) {
#define ZSCALE 1.35
	float r = T_RAD1 + T_RAD2*cos(2*M_PI*pos*Q);
	float theta = 2*M_PI*pos*P;
	float phi = 0.2*M_PI*sin(2*M_PI*pos*Q);
	vect[0] = F(r*cos(phi)*cos(theta));
	vect[1] = F(r*cos(phi)*sin(theta));
	vect[2] = F(ZSCALE*r*sin(phi));
}

bool hasExtension(const char *str) {
	const char *extensions = (const char*) glGetString(GL_EXTENSIONS);
	int len = strlen(str);
	const char *ptr;
	while ((ptr = strstr(extensions, str)) != NULL) {
		if ((ptr == extensions || *(ptr-1) == ' ') && (*(ptr+len) == ' ' || *(ptr+len) == '\0'))
			return true;
		extensions = ptr+len;
	}
	return false;
}

bool hasVersion(int major, int minor) {
	const char *version = (const char*) glGetString(GL_VERSION);
	while ((*version < '0' || *version > '9') && *version != '\0')
		version++;
	int v1, v2;
	if (sscanf(version, "%d.%d", &v1, &v2) != 2)
		return false;
	if (v1 < major) return false;
	if (v1 > major) return true;
	if (v2 < minor) return false;
	return true;
}

Knot::Knot(int w, int h) {
	width = w;
	height = h;
	updateBumps = false;
	motionBlur = false;
	radialBlur = false;
	morphing = false;

	texture = 0;
	celTexture = 0;
	motionTexture = 0;
	radialTexture = 0;

#ifndef NO_EGL
	if (hasExtension("OES_query_matrix") || hasExtension("GL_OES_query_matrix"))
		ext_glQueryMatrixxOES = (GLbitfield (*)(GLfixed[], GLint[])) eglGetProcAddress("glQueryMatrixxOES");
	else
#endif
		ext_glQueryMatrixxOES = NULL;

	precalc();

	reinit(UNTEXTURED);
}

void Knot::setUpdateVertices(bool update) {
	updateBumps = update;
	setTexMode(texMode);
	if (!updateBumps)
		calcDynVerticesNormals(vertices, normals, ringcoords);
}

void Knot::setMotionBlur(bool blur) {
	motionBlur = blur;
	setDrawState();
	if (blur) {
		glBindTexture(GL_TEXTURE_2D, motionTexture);
		GLubyte *data = new GLubyte[motionWidth*motionHeight*3];
		memset(data, 0, sizeof(GLubyte)*motionWidth*motionHeight*3);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, motionWidth, motionHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
		delete [] data;
	}
}

void Knot::setRadialBlur(bool blur) {
	radialBlur = blur;
	setDrawState();
}

void Knot::setRadialGlow(bool g) {
	glow = g;
}

void Knot::setSize(int w, int h) {
	width = w;
	height = h;
}

void Knot::reinit() {
	reinit(texMode);
}

void Knot::reinit(TexMode mode) {
	glViewport(0, 0, width, height);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslate(F(0), F(0), F(-10));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float zn = 5;
	float planeSize = zn*tan((54/2)*M_PI/180);
	float fw, fh;
	if (width < height) {
		fw = planeSize;
		fh = fw*height/width;
	} else {
		fh = planeSize;
		fw = fh*width/height;
	}
	glFrustum(F(-fw), F(fw), F(-fh), F(fh), F(zn), F(15));


	glDeleteTextures(1, &texture);
	glDeleteTextures(1, &celTexture);
	glDeleteTextures(1, &motionTexture);
	glDeleteTextures(1, &radialTexture);

	genTexture();
	genCelTexture();

	radialWidth = width;
	radialHeight = height;
	radialTexture = genEmptyTexture(radialWidth, radialHeight);
	motionWidth = width;
	motionHeight = height;
	motionTexture = genEmptyTexture(motionWidth, motionHeight);
	glBindTexture(GL_TEXTURE_2D, radialTexture);
//	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	lasttime = (unsigned int)-1;
	texOffsetX = texOffsetY = 0;
	bumpOffset = 0;

	setTexMode(mode);
}

static const TYPE lightPos[] = { F(-3), F(10), F(0), F(1) };
static const TYPE lightAmbient[] = { F(0.2), F(0.2), F(0.2), F(1) };
static const TYPE lightDiffuse[] = { F(1), F(1), F(1), F(1) };

void Knot::setTexMode(TexMode mode) {
#ifndef USE_OES_1_1
	if (mode == CELSHADED) {
		if (!ext_glQueryMatrixxOES)
			return;
	}
#endif

	glShadeModel(GL_SMOOTH);
	glClearDepth(F(1));
	glDepthRange(F(0), F(1));
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	static const TYPE lightSpecular[] = { F(1), F(1), F(1), F(1) };
	static const TYPE matSpecular[] = { F(0.8), F(0.8), F(0.8), F(1) };

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glLightv(GL_LIGHT0, GL_POSITION, lightPos);
	glPopMatrix();

	// if color material happens to be enabled, disable it
	// to make sure the material parameters actually are updated
	glDisable(GL_COLOR_MATERIAL);

	glEnable(GL_LIGHT0);
	glLightv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
	glLightv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
	glLightv(GL_LIGHT0, GL_SPECULAR, lightSpecular);
	glMaterialv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecular);
	glMaterial(GL_FRONT_AND_BACK, GL_SHININESS, F(35));

	static const TYPE ambient[] = { F(0.2), F(0.2), F(0.2), F(1) };
	static const TYPE diffuse[] = { F(0.8), F(0.8), F(0.8), F(1) };

	// disable the scene global ambient and enable ambient
	// on the light source, since n93 doesn't seem to use
	// the material ambient color with the scene global ambient
	static const TYPE sceneAmbient[] = { F(0), F(0), F(0), F(1) };
	glLightModelxv(GL_LIGHT_MODEL_AMBIENT, sceneAmbient);

	glMaterialv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
	glMaterialv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);

	texMode = mode;
	setDrawState();
}

void Knot::setDrawState() {
	glEnable(GL_LIGHTING);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_COLOR_MATERIAL);
	glClearColor(F(0), F(0), F(0), F(1));
	glColor4(F(1), F(1), F(1), F(1));
	glEnable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, texture);

	glVertexPointer(3, TYPETOKEN, 0, vertices);
	glNormalPointer(TYPETOKEN, 0, normals);
	glTexCoordPointer(2, TYPETOKEN, 0, texcoords);

	switch (texMode) {
	case UNTEXTURED:
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnable(GL_COLOR_MATERIAL);
		glColor4(F(0.1), F(0.1), F(1), F(1));
		break;

	case PLAINTEX:
	case TEXSPEC:
		glEnable(GL_TEXTURE_2D);
		break;

	case CELSHADED:
		glDisableClientState(GL_NORMAL_ARRAY);
		glEnable(GL_TEXTURE_2D);
		glClearColor(F(1), F(1), F(1), F(1));
		glDisable(GL_LIGHTING);
		glLineWidth(F(3));
		glBindTexture(GL_TEXTURE_2D, celTexture);
		glTexCoordPointer(2, TYPETOKEN, 0, celcoords);
		break;
	}
}

void Knot::genTexture() {
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	GLubyte* palette = new GLubyte[3*256];
	for (int i = 0; i < 256; i++) {
		GLubyte *b = &palette[3*i];
		b[0] = i;
		b[1] = i;
		b[2] = 255-i;
	}
	int size = 16;
	int level = 0;
	while (size >= 1) {
		GLubyte* pixels = new GLubyte[3*size*size];
		for (int y = 0; y < size; y++) {
			for (int x = 0; x < size; x++) {
				float fval = 0.5+0.5*sin(float(y)/size*2*M_PI + M_PI*sin(float(x)/size*2*M_PI));
				GLubyte val = GLubyte(255*fval);
				pixels[y*3*size + 3*x + 0] = palette[3*val + 0];
				pixels[y*3*size + 3*x + 1] = palette[3*val + 1];
				pixels[y*3*size + 3*x + 2] = palette[3*val + 2];
			}
		}
		glTexImage2D(GL_TEXTURE_2D, level, GL_RGB, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
		delete [] pixels;
		size /= 2;
		level++;
	}
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR);
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	delete [] palette;
}

void Knot::genCelTexture() {
	glGenTextures(1, &celTexture);
	glBindTexture(GL_TEXTURE_2D, celTexture);
	int steps = 2;
	static const float data[] = {
		1, 1, 1,
		0.60, 0.45, 0.45, 0.45,
		0.90, 0.20, 0.20, 0.20,
	};

	const float *dataptr = data;
	GLubyte r, g, b;
	r = int(255*(*dataptr++));
	g = int(255*(*dataptr++));
	b = int(255*(*dataptr++));

	float nextPos = 2;
	if (steps > 0)
		nextPos = *dataptr++;
	int next = 0;

	int size = 32;
	GLubyte* texdata = new GLubyte[3*size];
	for (int i = 0; i < size; i++) {
		float pos = float(i)/size;
		if (pos >= nextPos) {
			r = int(255*(*dataptr++));
			g = int(255*(*dataptr++));
			b = int(255*(*dataptr++));
			next++;
			if (next < steps)
				nextPos = *dataptr++;
			else
				nextPos = 2;
			
		}
		GLubyte *byte = &texdata[3*i];
		byte[0] = r;
		byte[1] = g;
		byte[2] = b;
	}
	int level = 0;
	while (size >= 1) {
		glTexImage2D(GL_TEXTURE_2D, level, GL_RGB, size, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, texdata);

		size /= 2;
		level++;
		if (size < 1) break;

		GLubyte *newtex = new GLubyte[3*size];
		for (int i = 0; i < size; i++) {
			GLubyte *out = &newtex[3*i];
			GLubyte *in = &texdata[2*3*i];
			out[0] = (int(in[0])+in[3])/2;
			out[1] = (int(in[1])+in[4])/2;
			out[2] = (int(in[2])+in[5])/2;
		}
		delete [] texdata;
		texdata = newtex;
	}
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	delete [] texdata;
}

static int largerPOT(int x) {
	int out = 1;
	while (out < x)
		out <<= 1;
	return out;
}
/*
static int smallerPOT(int x) {
	int out = largerPOT(x);
	if (out > x)
		out >>= 1;
	return out;
}
*/
GLuint Knot::genEmptyTexture(int &w, int &h) {
	GLuint tex;
	w = largerPOT(w);
	h = largerPOT(h);

	int max;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max);
	if (w > max) w = max;
	if (h > max) h = max;

	while (true) {
		glGenTextures(1, &tex);
		glBindTexture(GL_TEXTURE_2D, tex);
		GLubyte *data = new GLubyte[w*h*3];
		memset(data, 0, sizeof(GLubyte)*w*h*3);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
		delete [] data;
		int err = glGetError();
		if (err != GL_OUT_OF_MEMORY)
			break;

		glDeleteTextures(1, &tex);
		if (w >= h)
			w /= 2;
		else
			h /= 2;
	}
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterx(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	return tex;
}

void Knot::makeRingCoords(TYPE *coords, int P, int Q) {
	for (int i = 0; i < RINGS; i++) {
		TYPE *ring = &coords[16*i];
		TYPE *center = &ring[12];
		float pos = float(i)/RINGS;
		polar(pos, center, P, Q);
		TYPE *dir = &ring[8];
		polar(pos+float(1)/(5*RINGS), dir, P, Q);
		SUB3(dir, center);
		NORMALIZE3(dir);
		TYPE *xvect = &ring[0];
		TYPE *yvect = &ring[4];
		COPY3(xvect, center);
		xvect[2] = F(0);
		cross3(yvect, dir, xvect);
		cross3(xvect, yvect, dir);
		NORMALIZE3(xvect);
		NORMALIZE3(yvect);
		center[3] = F(1);
		dir[3] = F(0);
		xvect[3] = F(0);
		yvect[3] = F(0);
	}
}

void Knot::precalc() {
	for (int i = 0; i < SINETABLESIZE; i++)
		sinTable[i] = F(sin(2*M_PI*float(i)/SINETABLESIZE));

	makeRingCoords(ringcoords23, 2, 3);
	makeRingCoords(ringcoords25, 2, 5);
	ringcoords = ringcoords23;

	for (int i = 0; i < POINTS; i++) {
		TYPE* vertex = &ringvertices[4*i];
		float angle = 2*M_PI*i/POINTS;
		vertex[0] = F(cos(angle));
		vertex[1] = F(sin(angle));
		vertex[2] = F(1);
		vertex[3] = F(1);
	}

	for (int i = 0; i < RINGS; i++) {
		int cr = (POINTS+1)*i;
		int nr = (POINTS+1)*(i+1);
		for (int j = 0; j < POINTS; j++) {
			GLushort *offset = &indices[6*(POINTS*i+j)];
			int cp = j;
			int np = j+1;
			offset[0] = cr+cp;
			offset[1] = cr+np;
			offset[2] = nr+np;
			offset[3] = cr+cp;
			offset[4] = nr+np;
			offset[5] = nr+cp;
		}
	}

	bool bumps = updateBumps;
	updateBumps = false;
	calcDynVerticesNormals(vertices, normals, ringcoords);
	updateBumps = bumps;

#define MULTX 6
#define MULTY 100
	for (int i = 0; i <= RINGS; i++) {
		for (int j = 0; j <= POINTS; j++) {
			TYPE* texcoord = &texcoords[2*((POINTS+1)*i+j)];
			texcoord[0] = F(MULTX*float(j)/POINTS);
			texcoord[1] = F(MULTY*float(i)/RINGS);
		}
	}

	for (int i = 0; i <= RINGS; i++) {
		for (int j = 0; j <= POINTS; j++) {
			TYPE* coord = &celcoords[2*((POINTS+1)*i+j)];
			coord[0] = F(i)/RINGS;
			coord[1] = F(0);
		}
	}
}

Knot::~Knot() {
	glDeleteTextures(1, &texture);
	glDeleteTextures(1, &celTexture);
	glDeleteTextures(1, &motionTexture);
	glDeleteTextures(1, &radialTexture);
}

void Knot::startMorph() {
	if (morphing) return;
	morphing = true;
	morphStart = lasttime;
}

void Knot::draw(unsigned int msec) {
	if (lasttime == (unsigned int)-1)
		lasttime = msec;
	unsigned int diff = msec - lasttime;
	lasttime = msec;

	TYPE *verts = vertices;
	TYPE *norms = normals;

	if (morphing) {
#define MORPHLENGTH 1200
		if (msec - morphStart > MORPHLENGTH) {
			morphing = false;
			if (ringcoords == ringcoords23)
				ringcoords = ringcoords25;
			else
				ringcoords = ringcoords23;
			calcDynVerticesNormals(vertices, normals, ringcoords);
		} else {
			TYPE *next = ringcoords23;
			if (ringcoords == ringcoords23)
				next = ringcoords25;
			morphMatrices(ringcoordstemp, ringcoords, next, F(msec-morphStart)/MORPHLENGTH);
			calcDynVerticesNormals(verts, normals, ringcoordstemp);
		}
	} else if (updateBumps)
		calcDynVertices();

	int vpwidth = width;
	int vpheight = height;
	if (radialBlur || motionBlur)
		setDrawState();

#define MIN(a, b) ((a) < (b) ? (a) : (b))
	int motionOrigW = vpwidth, motionOrigH = vpheight;
	if (motionBlur) {
		if (motionWidth < vpwidth || motionHeight < vpheight) {
			vpwidth = MIN(motionWidth, vpwidth);
			vpheight = MIN(motionHeight, vpheight);
			glViewport(0, 0, vpwidth, vpheight);
		}
	}

	int radialOrigW = vpwidth, radialOrigH = vpheight;
	if (radialBlur) {
		if (radialWidth < vpwidth || radialHeight < vpheight) {
			vpwidth = MIN(radialWidth, vpwidth);
			vpheight = MIN(radialHeight, vpheight);
			glViewport(0, 0, vpwidth, vpheight);
		}
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glRotate(F(90)*diff/1000, F(sin(msec*M_PI/3000)), F(sin(msec*M_PI/600)), F(cos(msec*M_PI/900)));

	texOffsetY += F(diff)/30000;
	bumpOffset += F(diff)/4000;
#ifdef USING_FIXED
	texOffsetY &= 0x0000FFFF;
	bumpOffset &= 0x0000FFFF;
#else
	if (texOffsetY > 1) texOffsetY -= int(texOffsetY);
	if (bumpOffset > 1) bumpOffset -= int(bumpOffset);
#endif

	if (texMode == CELSHADED) {
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();

		updateOutlineVertices(outlineVertices, verts, norms);

		glCullFace(GL_FRONT);
		glColor4(F(0), F(0), F(0), F(1));
		glVertexPointer(3, TYPETOKEN, 0, outlineVertices);
		glDisable(GL_TEXTURE_2D);
		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);

		updateCelCoords(verts, norms);

		glCullFace(GL_BACK);
		glColor4(F(1), F(1), F(1), F(1));
		glVertexPointer(3, TYPETOKEN, 0, vertices);
		glEnable(GL_TEXTURE_2D);
		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);


	} else {

		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glTranslate(F(0), MULTY*texOffsetY, F(0));

		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);

		if (texMode == TEXSPEC) {
			static const TYPE zero[] = { F(0), F(0), F(0), F(0) };

			glEnable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glBlendFunc(GL_ONE, GL_ONE);
			glLightv(GL_LIGHT0, GL_DIFFUSE, zero);
			glLightv(GL_LIGHT0, GL_AMBIENT, zero);

			glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);

			glDisable(GL_BLEND);
			glEnable(GL_TEXTURE_2D);
			glLightv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
			glLightv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
		}
	}


	if (radialBlur) {
		glEnable(GL_TEXTURE_2D);
		glClearColor(F(0), F(0), F(0), F(1));
		glBindTexture(GL_TEXTURE_2D, radialTexture);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, vpwidth, vpheight);

		bool cleared = false;
		if (vpwidth < radialOrigW || vpheight < radialOrigH) {
			glViewport(0, 0, radialOrigW, radialOrigH);
			glClear(GL_COLOR_BUFFER_BIT);
			cleared = true;
		}

#define RADIALLAYERS 5

		TYPE tex[2*4*RADIALLAYERS];
		TYPE verts[2*4*RADIALLAYERS];
		TYPE colors[4*4*RADIALLAYERS];
		GLushort indices[6*RADIALLAYERS];
		static const GLushort rectindices[] = { 0, 1, 2, 0, 2, 3 };

		TYPE texWidth = F(vpwidth)/radialWidth;
		TYPE texHeight = F(vpheight)/radialHeight;
		for (int i = 0; i < RADIALLAYERS; i++) {
			verts[2*(4*i+0)+0] = F(0);
			verts[2*(4*i+0)+1] = F(0);
			verts[2*(4*i+1)+0] = F(100);
			verts[2*(4*i+1)+1] = F(0);
			verts[2*(4*i+2)+0] = F(100);
			verts[2*(4*i+2)+1] = F(100);
			verts[2*(4*i+3)+0] = F(0);
			verts[2*(4*i+3)+1] = F(100);
			for (int j = 0; j < 6; j++)
				indices[6*i+j] = rectindices[j] + 4*i;
			TYPE alpha = F(1)/(i+1);
			if (glow) {
				if (i == 0)
					alpha = F(1);
				else
					alpha = F(1)/RADIALLAYERS;
			} 
			for (int j = 0; j < 4; j++) {
				colors[4*(4*i+j)+0] = F(1);
				colors[4*(4*i+j)+1] = F(1);
				colors[4*(4*i+j)+2] = F(1);
				colors[4*(4*i+j)+3] = alpha;
			}
			TYPE margin = MULT(F(0.35), F(i))/RADIALLAYERS;
			TYPE xmargin = MULT(texWidth, margin)/2;
			TYPE ymargin = MULT(texHeight, margin)/2;
			tex[2*(4*i+0)+0] = xmargin;
			tex[2*(4*i+0)+1] = ymargin;
			tex[2*(4*i+1)+0] = texWidth - xmargin;
			tex[2*(4*i+1)+1] = ymargin;
			tex[2*(4*i+2)+0] = texWidth - xmargin;
			tex[2*(4*i+2)+1] = texHeight - ymargin;
			tex[2*(4*i+3)+0] = xmargin;
			tex[2*(4*i+3)+1] = texHeight - ymargin;
		}
		glColor4(F(1), F(1), F(1), F(1)/(RADIALLAYERS));


		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(F(0), F(100), F(0), F(100), F(-1), F(1));
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_COLOR_MATERIAL);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);

		glVertexPointer(2, TYPETOKEN, 0, verts);
		glTexCoordPointer(2, TYPETOKEN, 0, tex);
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);


		if (!cleared) {
			if (glow)
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			else
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glEnableClientState(GL_COLOR_ARRAY);
			glColorPointer(4, TYPETOKEN, 0, colors);
			glDrawElements(GL_TRIANGLES, 6*(RADIALLAYERS-1), GL_UNSIGNED_SHORT, indices+6);
		} else {
			if (glow) {
				glEnableClientState(GL_COLOR_ARRAY);
				glColorPointer(4, TYPETOKEN, 0, colors);
			}
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			glDrawElements(GL_TRIANGLES, 6*RADIALLAYERS, GL_UNSIGNED_SHORT, indices);
		}

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
		glDisableClientState(GL_COLOR_ARRAY);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		vpwidth = radialOrigW;
		vpheight = radialOrigH;
	}

	if (motionBlur) {
		TYPE alpha = F(1 - exp(-float(diff)/130));
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(F(0), F(100), F(0), F(100), F(-1), F(1));
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_COLOR_MATERIAL);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);

		glBindTexture(GL_TEXTURE_2D, motionTexture);
		glColor4(F(1), F(1), F(1), F(1) - alpha);

		TYPE tex[2*4];
		TYPE verts[2*4];
		tex[2*0+0] = F(0);
		tex[2*0+1] = F(0);
		tex[2*1+0] = F(vpwidth)/motionWidth;
		tex[2*1+1] = F(0);
		tex[2*2+0] = F(vpwidth)/motionWidth;
		tex[2*2+1] = F(vpheight)/motionHeight;
		tex[2*3+0] = F(0);
		tex[2*3+1] = F(vpheight)/motionHeight;
		verts[2*0+0] = F(0);
		verts[2*0+1] = F(0);
		verts[2*1+0] = F(100);
		verts[2*1+1] = F(0);
		verts[2*2+0] = F(100);
		verts[2*2+1] = F(100);
		verts[2*3+0] = F(0);
		verts[2*3+1] = F(100);
		static const GLushort indices[] = { 0, 1, 2, 0, 2, 3 };

		glVertexPointer(2, TYPETOKEN, 0, verts);
		glTexCoordPointer(2, TYPETOKEN, 0, tex);
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);

		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, vpwidth, vpheight);

		if (vpwidth < motionOrigW || vpheight < motionOrigH) {
			glViewport(0, 0, motionOrigW, motionOrigH);
			glDisable(GL_BLEND);
			glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);
		}

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		vpwidth = motionOrigW;
		vpheight = motionOrigH;
	}
}

void Knot::updateCelCoords(TYPE *verts, TYPE *norms) {
	TYPE light[4];
	getLightInverse(light);
	NORMALIZE3(light);
	TYPE *dir = light;

	for (int i = 0; i < RINGS; i++) {
//		TYPE* vertex = &verts[3*(POINTS+1)*i];
		TYPE* normal = &norms[3*(POINTS+1)*i];
		TYPE* coord = &celcoords[2*((POINTS+1)*i)];
		for (int j = 0; j < POINTS; j++) {
//			TYPE dir[4];
//			COPY3(dir, light);
//			SUB3(dir, vertex);
//			NORMALIZE3(dir);
			TYPE val = MULT(dir[0], normal[0]) + MULT(dir[1], normal[1]) + MULT(dir[2], normal[2]);
			if (val < F(0)) val = F(0);
			if (val > F(1)) val = F(1);
			coord[0] = F(1) - val;
			coord[1] = F(0);
//			vertex += 3;
			normal += 3;
			coord += 2;
		}
		memcpy(&celcoords[2*((POINTS+1)*i+POINTS)], &celcoords[2*((POINTS+1)*i+0)], sizeof(TYPE)*2);
	}
	memcpy(&celcoords[2*((POINTS+1)*RINGS)], &celcoords[0], sizeof(TYPE)*2*(POINTS+1));
}

void Knot::getLightInverse(TYPE* light) {

	TYPE rot[16], trans[16], lightMatrix[16];
	TYPE modelview[16];
	memset(rot, 0, sizeof(rot));
	memset(trans, 0, sizeof(rot));
	memset(lightMatrix, 0, sizeof(rot));

	glMatrixMode(GL_MODELVIEW);
#ifdef USE_OES_1_1
	glGetTypev(GL_MODELVIEW_MATRIX, modelview);
#else
	GLfixed mantissa[16];
	GLint exponent[16];
	ext_glQueryMatrixxOES(mantissa, exponent);
	for (int i = 0; i < 16; i++) {
		if (exponent[i] >= 0)
			modelview[i] = FIXED_TO_NATIVE(mantissa[i] << exponent[i]);
		else
			modelview[i] = FIXED_TO_NATIVE(mantissa[i] >> (-exponent[i]));
	}
#endif
	glPushMatrix();
	glLoadIdentity();

	// make rot and trans identity matrices
	for (int i = 0; i < 4; i++)
		rot[i+i*4] = trans[i+i*4] = F(1);

	// rot is the rotational part from modelview, transposed,
	// which also is the inverse of the rotational part
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			rot[i+j*4] = modelview[j+i*4];
	for (int i = 0; i < 3; i++)
		trans[i+3*4] = -modelview[i+3*4];
	for (int i = 0; i < 4; i++)
		lightMatrix[i] = lightPos[i];

	glMultMatrix(rot);
	glMultMatrix(trans);
	glMultMatrix(lightMatrix);

#ifdef USE_OES_1_1
	glGetTypev(GL_MODELVIEW_MATRIX, lightMatrix);
#else
	ext_glQueryMatrixxOES(mantissa, exponent);
	for (int i = 0; i < 16; i++) {
		if (exponent[i] >= 0)
			lightMatrix[i] = FIXED_TO_NATIVE(mantissa[i] << exponent[i]);
		else
			lightMatrix[i] = FIXED_TO_NATIVE(mantissa[i] >> (-exponent[i]));
	}
#endif
	glPopMatrix();

	for (int i = 0; i < 4; i++)
		light[i] = lightMatrix[i];
}

void Knot::calcDynVertices() {
	for (int i = 0; i < RINGS; i++) {
		const TYPE* ring = &ringcoords[16*i];
		const TYPE* center = &ring[12];
//		const TYPE* dir = &ring[8];
		const TYPE* xvect = &ring[0];
		const TYPE* yvect = &ring[4];
		TYPE pos = F(i)/RINGS - bumpOffset;
		TYPE radius = RADIUS + MULT(F(BUMPSIZE), fastSin(BUMPS*pos));

		for (int j = 0; j < POINTS; j++) {
			TYPE* vertex = &vertices[3*((POINTS+1)*i+j)];
			const TYPE* origvertex = &ringvertices[4*j];
			const TYPE &x = origvertex[0];
			const TYPE &y = origvertex[1];
			vertex[0] = center[0] + MULT(xvect[0], MULT(radius, x)) + MULT(yvect[0], MULT(radius, y));
			vertex[1] = center[1] + MULT(xvect[1], MULT(radius, x)) + MULT(yvect[1], MULT(radius, y));
			vertex[2] = center[2] + MULT(xvect[2], MULT(radius, x)) + MULT(yvect[2], MULT(radius, y));
//			vertex[3] = F(1);
		}
		memcpy(&vertices[3*((POINTS+1)*i+POINTS)], &vertices[3*((POINTS+1)*i+0)], sizeof(TYPE)*3);
	}
	memcpy(&vertices[3*((POINTS+1)*RINGS)], &vertices[0], sizeof(TYPE)*3*(POINTS+1));
}

void Knot::calcDynVerticesNormals(TYPE *verts, TYPE *norms, const TYPE *coords) {
	for (int i = 0; i < RINGS; i++) {
		const TYPE* ring = &coords[16*i];
		const TYPE* center = &ring[12];
//		const TYPE* dir = &ring[8];
		const TYPE* xvect = &ring[0];
		const TYPE* yvect = &ring[4];
		TYPE pos = F(i)/RINGS - bumpOffset;
		TYPE radius = RADIUS + MULT(F(BUMPSIZE), fastSin(BUMPS*pos));
		if (!updateBumps)
			radius = RADIUS;

		for (int j = 0; j < POINTS; j++) {
			TYPE* vertex = &verts[3*((POINTS+1)*i+j)];
			const TYPE* origvertex = &ringvertices[4*j];
			const TYPE &x = origvertex[0];
			const TYPE &y = origvertex[1];
			vertex[0] = center[0] + MULT(xvect[0], MULT(radius, x)) + MULT(yvect[0], MULT(radius, y));
			vertex[1] = center[1] + MULT(xvect[1], MULT(radius, x)) + MULT(yvect[1], MULT(radius, y));
			vertex[2] = center[2] + MULT(xvect[2], MULT(radius, x)) + MULT(yvect[2], MULT(radius, y));
//			vertex[3] = F(1);

			TYPE* normal = &norms[3*((POINTS+1)*i+j)];
			normal[0] = MULT(xvect[0], x) + MULT(yvect[0], y);
			normal[1] = MULT(xvect[1], x) + MULT(yvect[1], y);
			normal[2] = MULT(xvect[2], x) + MULT(yvect[2], y);
		}
		memcpy(&verts[3*((POINTS+1)*i+POINTS)], &verts[3*((POINTS+1)*i+0)], sizeof(TYPE)*3);
		memcpy(&norms[3*((POINTS+1)*i+POINTS)], &norms[3*((POINTS+1)*i+0)], sizeof(TYPE)*3);
	}
	memcpy(&verts[3*((POINTS+1)*RINGS)], &verts[0], sizeof(TYPE)*3*(POINTS+1));
	memcpy(&norms[3*((POINTS+1)*RINGS)], &norms[0], sizeof(TYPE)*3*(POINTS+1));
}

TYPE Knot::fastSin(TYPE x) {
#ifdef USING_FIXED
	return sinTable[(x >> (16-SINETABLEMASKSIZE)) & SINETABLEMASK];
#else
	return sinTable[int(SINETABLESIZE*x) & SINETABLEMASK];
#endif
}

void Knot::morphMatrices(TYPE *out, const TYPE *first, const TYPE *second, TYPE pos) {
	TYPE pos1 = F(1) - pos;
	for (int i = 0; i < RINGS; i++) {
		const TYPE* matrix1 = &first[16*i];
		const TYPE* center1 = &matrix1[12];
//		const TYPE* dir1 = &matrix1[8];
		const TYPE* xvect1 = &matrix1[0];
		const TYPE* yvect1 = &matrix1[4];
		const TYPE* matrix2 = &second[16*i];
		const TYPE* center2 = &matrix2[12];
//		const TYPE* dir2 = &matrix2[8];
		const TYPE* xvect2 = &matrix2[0];
		const TYPE* yvect2 = &matrix2[4];
		TYPE* matrix3 = &out[16*i];
		TYPE* center3 = &matrix3[12];
		TYPE* dir3 = &matrix3[8];
		TYPE* xvect3 = &matrix3[0];
		TYPE* yvect3 = &matrix3[4];
#define MULTVECT(vect3, vect1, fact1, vect2, fact2)					\
		do {									\
			vect3[0] = MULT(vect1[0], fact1) + MULT(vect2[0], fact2);	\
			vect3[1] = MULT(vect1[1], fact1) + MULT(vect2[1], fact2);	\
			vect3[2] = MULT(vect1[2], fact1) + MULT(vect2[2], fact2);	\
			vect3[3] = vect1[3];						\
		} while (0)
		MULTVECT(center3, center1, pos1, center2, pos);
		MULTVECT(xvect3, xvect1, pos1, xvect2, pos);
		MULTVECT(yvect3, yvect1, pos1, yvect2, pos);
		cross3(dir3, xvect3, yvect3);
		cross3(yvect3, dir3, xvect3);
		NORMALIZE3(xvect3);
		NORMALIZE3(yvect3);
		NORMALIZE3(dir3);
		dir3[3] = 0;
	}
}

void Knot::updateOutlineVertices(TYPE *outline, const TYPE *verts, const TYPE *norms) {
	int len = (RINGS + 1)*(POINTS + 1);
	for (int i = 0; i < len; i++) {
		const TYPE* vertex = &verts[3*i];
		const TYPE* normal = &norms[3*i];
		TYPE* out = &outline[3*i];
		out[0] = vertex[0] + MULT(F(OUTLINEADD), normal[0]);
		out[1] = vertex[1] + MULT(F(OUTLINEADD), normal[1]);
		out[2] = vertex[2] + MULT(F(OUTLINEADD), normal[2]);
	}
}



